@@ -137,7 +137,7 @@ static inline int pthread_cancel (pthread_t thread) {
 }
 
 static inline void pthread_exit (void *res) {
-  ExitThread((DWORD_PTR) (DWORD) res);
+  ExitThread((DWORD) (DWORD_PTR) res);
 }
 
 static inline int pthread_join (pthread_t thread, void **res) {
