@@ -1000,11 +1000,6 @@ double _starpu_history_based_job_expected_perf(struct starpu_perfmodel *model, e
 
 	_STARPU_PTHREAD_RWLOCK_RDLOCK(&model->model_rwlock);
 	history = per_arch_model->history;
-	if (!history) {
-		_STARPU_PTHREAD_RWLOCK_UNLOCK(&model->model_rwlock);
-		return NAN;
-	}
-
 	entry = (struct starpu_history_entry *) _starpu_htbl_search_32(history, key);
 	_STARPU_PTHREAD_RWLOCK_UNLOCK(&model->model_rwlock);
 
@@ -1103,16 +1098,17 @@ void _starpu_update_perfmodel_history(struct _starpu_job *j, struct starpu_perfm
 				reg_model->maxx = job_size;
 			reg_model->nsample++;
 
-			unsigned n = reg_model->nsample;
-
-			double num = (n*reg_model->sumlnxlny - reg_model->sumlnx*reg_model->sumlny);
-			double denom = (n*reg_model->sumlnx2 - reg_model->sumlnx*reg_model->sumlnx);
-
-			reg_model->beta = num/denom;
-			reg_model->alpha = exp((reg_model->sumlny - reg_model->beta*reg_model->sumlnx)/n);
-
 			if (VALID_REGRESSION(reg_model))
+			{
+				unsigned n = reg_model->nsample;
+
+				double num = (n*reg_model->sumlnxlny - reg_model->sumlnx*reg_model->sumlny);
+				double denom = (n*reg_model->sumlnx2 - reg_model->sumlnx*reg_model->sumlnx);
+
+				reg_model->beta = num/denom;
+				reg_model->alpha = exp((reg_model->sumlny - reg_model->beta*reg_model->sumlnx)/n);
 				reg_model->valid = 1;
+			}
 		}
 
 #ifdef STARPU_MODEL_DEBUG
