@@ -1,8 +1,9 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
- * Copyright (C) 2010-2011  Université de Bordeaux 1
+ * Copyright (C) 2010-2012  Université de Bordeaux 1
  * Copyright (C) 2010  Mehdi Juhoor <mjuhoor@gmail.com>
- * Copyright (C) 2010, 2011  Centre National de la Recherche Scientifique
+ * Copyright (C) 2010, 2011, 2012  Centre National de la Recherche Scientifique
+ * Copyright (C) 2012 INRIA
  *
  * StarPU is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
@@ -19,12 +20,12 @@
 #include <datawizard/filters.h>
 #include <datawizard/footprint.h>
 
-static void starpu_data_create_children(starpu_data_handle handle, unsigned nchildren, struct starpu_data_filter *f);
+static void starpu_data_create_children(starpu_data_handle_t handle, unsigned nchildren, struct starpu_data_filter *f);
 
 /*
  * This function applies a data filter on all the elements of a partition
  */
-static void map_filter(starpu_data_handle root_handle, struct starpu_data_filter *f)
+static void map_filter(starpu_data_handle_t root_handle, struct starpu_data_filter *f)
 {
 	/* we need to apply the data filter on all leaf of the tree */
 	if (root_handle->nchildren == 0)
@@ -32,7 +33,8 @@ static void map_filter(starpu_data_handle root_handle, struct starpu_data_filter
 		/* this is a leaf */
 		starpu_data_partition(root_handle, f);
 	}
-	else {
+	else
+	{
 		/* try to apply the data filter recursively */
 		unsigned child;
 		for (child = 0; child < root_handle->nchildren; child++)
@@ -41,11 +43,9 @@ static void map_filter(starpu_data_handle root_handle, struct starpu_data_filter
 		}
 	}
 }
-void starpu_data_map_filters(starpu_data_handle root_handle, unsigned nfilters, ...)
+void starpu_data_vmap_filters(starpu_data_handle_t root_handle, unsigned nfilters, va_list pa)
 {
 	unsigned i;
-	va_list pa;
-	va_start(pa, nfilters);
 	for (i = 0; i < nfilters; i++)
 	{
 		struct starpu_data_filter *next_filter;
@@ -55,15 +55,22 @@ void starpu_data_map_filters(starpu_data_handle root_handle, unsigned nfilters,
 
 		map_filter(root_handle, next_filter);
 	}
+}
+
+void starpu_data_map_filters(starpu_data_handle_t root_handle, unsigned nfilters, ...)
+{
+	va_list pa;
+	va_start(pa, nfilters);
+	starpu_data_vmap_filters(root_handle, nfilters, pa);
 	va_end(pa);
 }
 
-int starpu_data_get_nb_children(starpu_data_handle handle)
+int starpu_data_get_nb_children(starpu_data_handle_t handle)
 {
         return handle->nchildren;
 }
 
-starpu_data_handle starpu_data_get_child(starpu_data_handle handle, unsigned i)
+starpu_data_handle_t starpu_data_get_child(starpu_data_handle_t handle, unsigned i)
 {
 	STARPU_ASSERT(i < handle->nchildren);
 
@@ -71,17 +78,25 @@ starpu_data_handle starpu_data_get_child(starpu_data_handle handle, unsigned i)
 }
 
 /*
- * example starpu_data_get_sub_data(starpu_data_handle root_handle, 3, 42, 0, 1);
+ * example starpu_data_get_sub_data(starpu_data_handle_t root_handle, 3, 42, 0, 1);
  */
-starpu_data_handle starpu_data_get_sub_data(starpu_data_handle root_handle, unsigned depth, ... )
+starpu_data_handle_t starpu_data_get_sub_data(starpu_data_handle_t root_handle, unsigned depth, ... )
 {
-	STARPU_ASSERT(root_handle);
-	starpu_data_handle current_handle = root_handle;
-
-	/* the variable number of argument must correlate the depth in the tree */
-	unsigned i; 
 	va_list pa;
 	va_start(pa, depth);
+	starpu_data_handle_t handle = starpu_data_vget_sub_data(root_handle, depth, pa);
+	va_end(pa);
+
+	return handle;
+}
+
+starpu_data_handle_t starpu_data_vget_sub_data(starpu_data_handle_t root_handle, unsigned depth, va_list pa )
+{
+	STARPU_ASSERT(root_handle);
+	starpu_data_handle_t current_handle = root_handle;
+
+	/* the variable number of argument must correlate the depth in the tree */
+	unsigned i;
 	for (i = 0; i < depth; i++)
 	{
 		unsigned next_child;
@@ -91,21 +106,20 @@ starpu_data_handle starpu_data_get_sub_data(starpu_data_handle root_handle, unsi
 
 		current_handle = &current_handle->children[next_child];
 	}
-	va_end(pa);
 
 	return current_handle;
 }
 
-void starpu_data_partition(starpu_data_handle initial_handle, struct starpu_data_filter *f)
+void starpu_data_partition(starpu_data_handle_t initial_handle, struct starpu_data_filter *f)
 {
 	unsigned nparts;
 	unsigned i;
+	unsigned node;
 
 	/* first take care to properly lock the data header */
 	_starpu_spin_lock(&initial_handle->header_lock);
 
-	/* there should not be mutiple filters applied on the same data */
-	STARPU_ASSERT(initial_handle->nchildren == 0);
+	STARPU_ASSERT_MSG(initial_handle->nchildren == 0, "there should not be mutiple filters applied on the same data");
 
 	/* how many parts ? */
 	if (f->get_nchildren)
@@ -120,9 +134,21 @@ void starpu_data_partition(starpu_data_handle initial_handle, struct starpu_data
 
 	unsigned nworkers = starpu_worker_get_count();
 
+	for (node = 0; node < STARPU_MAXNODES; node++)
+	{
+		if (initial_handle->per_node[node].state != STARPU_INVALID)
+			break;
+	}
+	if (node == STARPU_MAXNODES) {
+		/* This is lazy allocation, allocate it now in main RAM, so as
+		 * to have somewhere to gather pieces later */
+		int ret = _starpu_allocate_memory_on_node(initial_handle, &initial_handle->per_node[0], 0);
+		STARPU_ASSERT(!ret);
+	}
+
 	for (i = 0; i < nparts; i++)
 	{
-		starpu_data_handle child =
+		starpu_data_handle_t child =
 			starpu_data_get_child(initial_handle, i);
 
 		STARPU_ASSERT(child);
@@ -140,14 +166,19 @@ void starpu_data_partition(starpu_data_handle initial_handle, struct starpu_data
 		child->is_readonly = initial_handle->is_readonly;
 
 		/* initialize the chunk lock */
-		child->req_list = starpu_data_requester_list_new();
-		child->reduction_req_list = starpu_data_requester_list_new();
+		child->req_list = _starpu_data_requester_list_new();
+		child->reduction_req_list = _starpu_data_requester_list_new();
 		child->refcnt = 0;
+		child->busy_count = 0;
+		child->busy_waiting = 0;
+		_STARPU_PTHREAD_MUTEX_INIT(&child->busy_mutex, NULL);
+		_STARPU_PTHREAD_COND_INIT(&child->busy_cond, NULL);
+		child->reduction_refcnt = 0;
 		_starpu_spin_init(&child->header_lock);
 
 		child->sequential_consistency = initial_handle->sequential_consistency;
 
-		PTHREAD_MUTEX_INIT(&child->sequential_consistency_mutex, NULL);
+		_STARPU_PTHREAD_MUTEX_INIT(&child->sequential_consistency_mutex, NULL);
 		child->last_submitted_mode = STARPU_R;
 		child->last_submitted_writer = NULL;
 		child->last_submitted_readers = NULL;
@@ -159,20 +190,16 @@ void starpu_data_partition(starpu_data_handle initial_handle, struct starpu_data
 		child->redux_cl = initial_handle->redux_cl;
 		child->init_cl = initial_handle->init_cl;
 
-		child->reduction_refcnt = 0;
-		child->reduction_req_list = starpu_data_requester_list_new();
-
 #ifdef STARPU_USE_FXT
 		child->last_submitted_ghost_writer_id_is_valid = 0;
 		child->last_submitted_ghost_writer_id = 0;
 		child->last_submitted_ghost_readers_id = NULL;
 #endif
 
-		unsigned node;
 		for (node = 0; node < STARPU_MAXNODES; node++)
 		{
-			struct starpu_data_replicate_s *initial_replicate; 
-			struct starpu_data_replicate_s *child_replicate;
+			struct _starpu_data_replicate *initial_replicate;
+			struct _starpu_data_replicate *child_replicate;
 
 			initial_replicate = &initial_handle->per_node[node];
 			child_replicate = &child->per_node[node];
@@ -183,7 +210,7 @@ void starpu_data_partition(starpu_data_handle initial_handle, struct starpu_data
 			child_replicate->refcnt = 0;
 			child_replicate->memory_node = node;
 			child_replicate->relaxed_coherency = 0;
-			
+
 			/* update the interface */
 			void *initial_interface = starpu_data_get_interface_on_node(initial_handle, node);
 			void *child_interface = starpu_data_get_interface_on_node(child, node);
@@ -194,9 +221,9 @@ void starpu_data_partition(starpu_data_handle initial_handle, struct starpu_data
 		unsigned worker;
 		for (worker = 0; worker < nworkers; worker++)
 		{
-			struct starpu_data_replicate_s *child_replicate;
+			struct _starpu_data_replicate *child_replicate;
 			child_replicate = &child->per_worker[worker];
-			
+
 			child_replicate->state = STARPU_INVALID;
 			child_replicate->allocated = 0;
 			child_replicate->automatically_allocated = 0;
@@ -232,37 +259,72 @@ void starpu_data_partition(starpu_data_handle initial_handle, struct starpu_data
 	_starpu_spin_unlock(&initial_handle->header_lock);
 }
 
-void starpu_data_unpartition(starpu_data_handle root_handle, uint32_t gathering_node)
+static
+void _starpu_empty_codelet_function(void *buffers[], void *args)
+{
+	(void) buffers; // unused;
+	(void) args; // unused;
+}
+
+void starpu_data_unpartition(starpu_data_handle_t root_handle, uint32_t gathering_node)
 {
 	unsigned child;
 	unsigned node;
 
 	_starpu_spin_lock(&root_handle->header_lock);
 
+	STARPU_ASSERT_MSG(root_handle->nchildren != 0, "data is not partitioned");
+
 	/* first take all the children lock (in order !) */
 	for (child = 0; child < root_handle->nchildren; child++)
 	{
-		struct starpu_data_state_t *child_handle = &root_handle->children[child];
+		struct _starpu_data_state *child_handle = &root_handle->children[child];
 
 		/* make sure the intermediate children is unpartitionned as well */
 		if (child_handle->nchildren > 0)
 			starpu_data_unpartition(child_handle, gathering_node);
 
+		/* If this is a multiformat handle, we must convert the data now */
+#ifdef STARPU_DEVEL
+#warning TODO: _starpu_fetch_data_on_node should be doing it
+#endif
+		if (_starpu_data_is_multiformat_handle(child_handle) &&
+			starpu_node_get_kind(child_handle->mf_node) != STARPU_CPU_RAM)
+		{
+			struct starpu_codelet cl =
+			{
+				.where = STARPU_CPU,
+				.cpu_funcs = { _starpu_empty_codelet_function, NULL },
+				.modes = { STARPU_RW },
+				.nbuffers = 1
+			};
+			struct starpu_task *task = starpu_task_create();
+			task->handles[0] = child_handle;
+			task->cl = &cl;
+			task->synchronous = 1;
+			if (starpu_task_submit(task) != 0)
+				_STARPU_ERROR("Could not submit the conversion task while unpartitionning\n");
+		}
+
 		int ret;
-		ret = _starpu_fetch_data_on_node(child_handle, &child_handle->per_node[gathering_node], STARPU_R, 0, NULL, NULL);
-		/* for now we pretend that the RAM is almost unlimited and that gathering 
+		ret = _starpu_fetch_data_on_node(child_handle, &child_handle->per_node[gathering_node], STARPU_R, 0, 0, NULL, NULL);
+		/* for now we pretend that the RAM is almost unlimited and that gathering
 		 * data should be possible from the node that does the unpartionning ... we
 		 * don't want to have the programming deal with memory shortage at that time,
 		 * really */
-		STARPU_ASSERT(ret == 0); 
+		STARPU_ASSERT(ret == 0);
+
+		_starpu_spin_lock(&child_handle->header_lock);
 
 		_starpu_data_free_interfaces(&root_handle->children[child]);
+		_starpu_data_requester_list_delete(child_handle->req_list);
+		_starpu_data_requester_list_delete(child_handle->reduction_req_list);
 	}
 
 	/* the gathering_node should now have a valid copy of all the children.
 	 * For all nodes, if the node had all copies and none was locally
 	 * allocated then the data is still valid there, else, it's invalidated
-	 * for the gathering node, if we have some locally allocated data, we 
+	 * for the gathering node, if we have some locally allocated data, we
 	 * copy all the children (XXX this should not happen so we just do not
 	 * do anything since this is transparent ?) */
 	unsigned still_valid[STARPU_MAXNODES];
@@ -281,40 +343,54 @@ void starpu_data_unpartition(starpu_data_handle root_handle, uint32_t gathering_
 
 		for (child = 0; child < root_handle->nchildren; child++)
 		{
-			struct starpu_data_replicate_s *local = &root_handle->children[child].per_node[node];
+			struct _starpu_data_replicate *local = &root_handle->children[child].per_node[node];
 
-			if (local->state == STARPU_INVALID) {
-				isvalid = 0; 
+			if (local->state == STARPU_INVALID)
+			{
+				/* One of the bits is missing */
+				isvalid = 0;
 			}
-	
-			if (local->allocated && local->automatically_allocated){
-				/* free the data copy in a lazy fashion */
-				_starpu_request_mem_chunk_removal(root_handle, node);
-				isvalid = 0; 
-			}
-#ifdef STARPU_DEVEL
-#warning free the data replicate if needed
-#endif
 
+			if (local->allocated && local->automatically_allocated)
+				/* free the child data copy in a lazy fashion */
+				_starpu_request_mem_chunk_removal(&root_handle->children[child], node, 1);
 		}
 
+		if (!root_handle->per_node[node].allocated)
+			/* Even if we have all the bits, if we don't have the
+			 * whole data, it's not valid */
+			isvalid = 0;
+
+		if (!isvalid && root_handle->per_node[node].allocated && root_handle->per_node[node].automatically_allocated)
+			/* free the data copy in a lazy fashion */
+			_starpu_request_mem_chunk_removal(root_handle, node, 1);
+
 		/* if there was no invalid copy, the node still has a valid copy */
 		still_valid[node] = isvalid;
-		nvalids++;
+		if (isvalid)
+			nvalids++;
 	}
 
 	/* either shared or owned */
 	STARPU_ASSERT(nvalids > 0);
 
-	starpu_cache_state newstate = (nvalids == 1)?STARPU_OWNER:STARPU_SHARED;
+	enum _starpu_cache_state newstate = (nvalids == 1)?STARPU_OWNER:STARPU_SHARED;
 
 	for (node = 0; node < STARPU_MAXNODES; node++)
 	{
-		root_handle->per_node[node].state = 
+		root_handle->per_node[node].state =
 			still_valid[node]?newstate:STARPU_INVALID;
 	}
 
+	for (child = 0; child < root_handle->nchildren; child++)
+	{
+		struct _starpu_data_state *child_handle = &root_handle->children[child];
+		_starpu_spin_unlock(&child_handle->header_lock);
+	}
+
 	/* there is no child anymore */
+	free(root_handle->children);
+	root_handle->children = NULL;
 	root_handle->nchildren = 0;
 
 	/* now the parent may be used again so we release the lock */
@@ -322,9 +398,9 @@ void starpu_data_unpartition(starpu_data_handle root_handle, uint32_t gathering_
 }
 
 /* each child may have his own interface type */
-static void starpu_data_create_children(starpu_data_handle handle, unsigned nchildren, struct starpu_data_filter *f)
+static void starpu_data_create_children(starpu_data_handle_t handle, unsigned nchildren, struct starpu_data_filter *f)
 {
-	handle->children = calloc(nchildren, sizeof(struct starpu_data_state_t));
+	handle->children = (struct _starpu_data_state *) calloc(nchildren, sizeof(struct _starpu_data_state));
 	STARPU_ASSERT(handle->children);
 
 	unsigned node;
@@ -335,16 +411,16 @@ static void starpu_data_create_children(starpu_data_handle handle, unsigned nchi
 
 	for (child = 0; child < nchildren; child++)
 	{
-		starpu_data_handle handle_child = &handle->children[child];
-		
-		struct starpu_data_interface_ops_t *ops;
-		
+		starpu_data_handle_t handle_child = &handle->children[child];
+
+		struct starpu_data_interface_ops *ops;
+
 		/* what's this child's interface ? */
 		if (f->get_child_ops)
 		  ops = f->get_child_ops(f, child);
 		else
 		  ops = handle->ops;
-		
+
 		handle_child->ops = ops;
 
 		size_t interfacesize = ops->interface_size;
@@ -363,8 +439,10 @@ static void starpu_data_create_children(starpu_data_handle handle, unsigned nchi
 			handle_child->per_worker[worker].data_interface = calloc(1, interfacesize);
 			STARPU_ASSERT(handle_child->per_worker[worker].data_interface);
 		}
+
+		handle_child->mf_node = handle->mf_node;
 	}
-	
+
 	/* this handle now has children */
 	handle->nchildren = nchildren;
 }
