@@ -27,10 +27,10 @@
 #include <stdlib.h>
 #include "../helper.h"
 
-#ifdef STARPU_SLOW_MACHINE
-#define NLOOPS		100
+#ifdef STARPU_QUICK_CHECK
+#  define NLOOPS		100
 #else
-#define NLOOPS		1000
+#  define NLOOPS		1000
 #endif
 #define VECTORSIZE	1024
 
