@@ -1,7 +1,8 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
- * Copyright (C) 2010  Université de Bordeaux 1
- * Copyright (C) 2010, 2011  Centre National de la Recherche Scientifique
+ * Copyright (C) 2010, 2012  Université de Bordeaux 1
+ * Copyright (C) 2010, 2011, 2012  Centre National de la Recherche Scientifique
+ * Copyright (C) 2012 inria
  *
  * StarPU is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
@@ -15,6 +16,7 @@
  * See the GNU Lesser General Public License in COPYING.LGPL for more details.
  */
 
+#include <config.h>
 #include <stdio.h>
 #include <unistd.h>
 #include <errno.h>
@@ -24,9 +26,13 @@
 #include <gordon.h>
 #endif
 
-#define N	100
-#define K	256
-#define FPRINTF(ofile, fmt, args ...) do { if (!getenv("STARPU_SSILENT")) {fprintf(ofile, fmt, ##args); }} while(0)
+#include "../helper.h"
+
+#define N_DEF	100
+#define K_DEF	256
+
+static int n=N_DEF;
+static int k=K_DEF;
 
 /*
  * In this test, we maintain a vector v = (a,b,c).
@@ -57,24 +63,37 @@ struct starpu_opencl_program opencl_code;
 
 #define VECTORSIZE	16
 
-starpu_data_handle v_handle;
+starpu_data_handle_t v_handle;
 static unsigned v[VECTORSIZE] __attribute__((aligned(128))) = {0, 0, 0, 0};
 
 void cpu_codelet_incA(void *descr[], __attribute__ ((unused)) void *_args)
 {
+	STARPU_SKIP_IF_VALGRIND;
+
 	unsigned *val = (unsigned *)STARPU_VECTOR_GET_PTR(descr[0]);
 	val[0]++;
 }
 
 void cpu_codelet_incC(void *descr[], __attribute__ ((unused)) void *_args)
 {
+	STARPU_SKIP_IF_VALGRIND;
+
 	unsigned *val = (unsigned *)STARPU_VECTOR_GET_PTR(descr[0]);
 	val[2]++;
 }
 
 int main(int argc, char **argv)
 {
-	starpu_init(NULL);
+	int ret;
+
+#ifdef STARPU_QUICK_CHECK
+	n /= 10;
+	k /= 16;
+#endif
+
+	ret = starpu_init(NULL);
+	if (ret == -ENODEV) return STARPU_TEST_SKIPPED;
+	STARPU_CHECK_RETURN_VALUE(ret, "starpu_init");
 
 #ifdef STARPU_USE_GORDON
 	unsigned elf_id = gordon_register_elf_plugin("./datawizard/sync_and_notify_data_gordon_kernels.spuelf");
@@ -90,104 +109,115 @@ int main(int argc, char **argv)
 #endif
 
 #ifdef STARPU_USE_OPENCL
-        starpu_opencl_load_opencl_from_file("tests/datawizard/sync_and_notify_data_opencl_codelet.cl", &opencl_code, NULL);
+        ret = starpu_opencl_load_opencl_from_file("tests/datawizard/sync_and_notify_data_opencl_codelet.cl", &opencl_code, NULL);
+	STARPU_CHECK_RETURN_VALUE(ret, "starpu_opencl_load_opencl_from_file");
 #endif
 
         starpu_vector_data_register(&v_handle, 0, (uintptr_t)v, VECTORSIZE, sizeof(unsigned));
 
 	unsigned iter;
-	for (iter = 0; iter < K; iter++)
+	for (iter = 0; iter < k; iter++)
 	{
 		int ret;
 		unsigned ind;
-		for (ind = 0; ind < N; ind++)
+		for (ind = 0; ind < n; ind++)
 		{
 			/* increment a = v[0] */
-			starpu_codelet cl_inc_a = {
+			struct starpu_codelet cl_inc_a =
+			{
 				.where = STARPU_CPU|STARPU_CUDA|STARPU_OPENCL|STARPU_GORDON,
-				.cpu_func = cpu_codelet_incA,
+				.cpu_funcs = {cpu_codelet_incA, NULL},
 #ifdef STARPU_USE_CUDA
-				.cuda_func = cuda_codelet_incA,
+				.cuda_funcs = {cuda_codelet_incA, NULL},
 #endif
 #ifdef STARPU_USE_OPENCL
-				.opencl_func = opencl_codelet_incA,
+				.opencl_funcs = {opencl_codelet_incA, NULL},
 #endif
 #ifdef STARPU_USE_GORDON
 				.gordon_func = kernel_incA_id,
 #endif
-				.nbuffers = 1
+				.nbuffers = 1,
+				.modes = {STARPU_RW}
 			};
 
 			struct starpu_task *task = starpu_task_create();
 			task->cl = &cl_inc_a;
-
-			task->buffers[0].handle = v_handle;
-			task->buffers[0].mode = STARPU_RW;
+			task->handles[0] = v_handle;
 
 			task->synchronous = 1;
 
 			ret = starpu_task_submit(task);
-			if (ret == -ENODEV)
-				goto enodev;
+			if (ret == -ENODEV) goto enodev;
+			STARPU_CHECK_RETURN_VALUE(ret, "starpu_task_submit");
 		}
 
 		/* synchronize v in RAM */
-		starpu_data_acquire(v_handle, STARPU_RW);
+		ret = starpu_data_acquire(v_handle, STARPU_RW);
+		STARPU_CHECK_RETURN_VALUE(ret, "starpu_data_acquire");
 
 		/* increment b */
 		v[1]++;
 
 		starpu_data_release(v_handle);
 
-		for (ind = 0; ind < N; ind++)
+		for (ind = 0; ind < n; ind++)
 		{
 			/* increment c = v[2] */
-			starpu_codelet cl_inc_c = {
+			struct starpu_codelet cl_inc_c =
+			{
 				.where = STARPU_CPU|STARPU_CUDA|STARPU_OPENCL|STARPU_GORDON,
-				.cpu_func = cpu_codelet_incC,
+				.cpu_funcs = {cpu_codelet_incC, NULL},
 #ifdef STARPU_USE_CUDA
-				.cuda_func = cuda_codelet_incC,
+				.cuda_funcs = {cuda_codelet_incC, NULL},
 #endif
 #ifdef STARPU_USE_OPENCL
-				.opencl_func = opencl_codelet_incC,
+				.opencl_funcs = {opencl_codelet_incC, NULL},
 #endif
 #ifdef STARPU_USE_GORDON
 				.gordon_func = kernel_incC_id,
 #endif
-				.nbuffers = 1
+				.nbuffers = 1,
+				.modes = {STARPU_RW}
+
 			};
 
 			struct starpu_task *task = starpu_task_create();
 			task->cl = &cl_inc_c;
 
-			task->buffers[0].handle = v_handle;
-			task->buffers[0].mode = STARPU_RW;
+			task->handles[0] = v_handle;
 
 			task->synchronous = 1;
 
 			ret = starpu_task_submit(task);
-			if (ret == -ENODEV)
-				goto enodev;
+			if (ret == -ENODEV) goto enodev;
+			STARPU_CHECK_RETURN_VALUE(ret, "starpu_task_submit");
 		}
 
 	}
 
-	starpu_data_acquire(v_handle, STARPU_RW);
+	ret = starpu_data_acquire(v_handle, STARPU_RW);
+	STARPU_CHECK_RETURN_VALUE(ret, "starpu_data_acquire");
 
 	FPRINTF(stderr, "V = {%u, %u, %u}\n", v[0], v[1], v[2]);
 
 	starpu_data_release(v_handle);
+	starpu_data_unregister(v_handle);
 
 	starpu_shutdown();
 
-	if ((v[0] != N*K) || (v[1] != K) || (v[2] != N*K))
-		return -1;
+	if ((v[0] != n*k) || (v[1] != k) || (v[2] != n*k))
+	{
+		FPRINTF(stderr, "Incorrect result\n");
+		STARPU_RETURN(EXIT_FAILURE);
+	}
 
-	return 0;
+	STARPU_RETURN(EXIT_SUCCESS);
 
 enodev:
+	starpu_data_unregister(v_handle);
+	starpu_shutdown();
 	fprintf(stderr, "WARNING: No one can execute this task\n");
 	/* yes, we do not perform the computation but we did detect that no one
  	 * could perform the kernel, so this is not an error from StarPU */
-	return 77;
+	STARPU_RETURN(STARPU_TEST_SKIPPED);
 }
